// This code is copyright Joe Kopena (scarius@geocities.com; UIN 1783204)
// If you like it, tell me about it. If you can improve it, tell me about it
// I'm not responsible for anything that happens to you while using this program,
// code or anything I ever made anytime.

#include <stdio.h>
#include <allegro.h>
#include <noflick.h>

#define pIndex(x, y)    ((y*bitmap->w)+x)

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
int snagSave(BITMAP *bmp);

int bSnag(int lx, int ty, int rx, int by);
int boxGrab(int x, int y);

void cgProcess(int x, int y);
int clickGrab(int x, int y);

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
char filePrefix[5];
char fileType[4];
BITMAP *bitmap;
PALETTE pal;
char *processed;
int minx, miny, maxx, maxy; // These were made global so they wouldn't have to be
                // passed during the click grab processes.

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
int main(int argc, char *argv[]) {
int err = 0, done=0;

   if (argc < 2) {
      printf("Usage: snag bitmap [result prefix] [result type]\n");
      printf("       If you want to specify a new file type (default is bmp)\n");
      printf("       you must specify a prefix, in the order above.\n");
      return -1;
   }

   if (argc >= 3)
      strcpy(filePrefix, argv[2]);
   else strcpy(filePrefix, "res");

   if (argc >= 4)
      strcpy(fileType, argv[3]);
   else strcpy(fileType, "bmp");

//   if (strstr(filename, ".bmp") || strstr(filename, ".pcx")
//      || !strstr(filename, ".tga")) {

   allegro_init();
   if (install_mouse() < 0) {
      fprintf(stderr, "Mouse error!\n");
      return -1;
   }
   install_timer();
   install_keyboard();
   clear_keybuf();

   if ((err = set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0))) {
      fprintf(stderr, "Set gfx mode error\n");
      return err;
   }

   if ((err = ff_setup(SCREEN_W, SCREEN_H))) {
      fprintf(stderr, "noflick error\n");
      return err;
   }

   bitmap = load_bitmap(argv[1], pal);
   if (!bitmap) {
      fprintf(stderr, "Could not read bitmap\n");
      return -2;
   }

   processed = malloc(sizeof(char) * (bitmap->w * bitmap->h));
   if (!processed) {
      printf("There was not enough memory for the table of processed elements\n");
      return -4;
   }

   position_mouse(bitmap->w / 2, bitmap->h / 2);
   set_mouse_range(0, 0, bitmap->w-1, bitmap->h-1);

   set_palette(pal);
   blit(bitmap, screen, 0, 0, 0, 0, bitmap->w, bitmap->h);

   show_mouse(screen);

   while (!done && !err) {
         while (!mouse_b && !keypressed()) {
//               fprintf(stderr, "x, y: (%i, %i)\n", mouse_x, mouse_y);
         }

         if (keypressed()) {
            if (key[KEY_ESC])
               done = 1;
         } else {
           if (mouse_b & 2) // Right button
              err = boxGrab(mouse_x, mouse_y);
           else if (mouse_b & 1) // left button
                err = clickGrab(mouse_x, mouse_y);

           while (mouse_b)
                 ; // Let go of the damn thing already!!!!
         } // Musta been a mouse clicky

   } // !done && !err

   free(processed);
   destroy_bitmap(bitmap);

   ff_shutdown();
   clear_keybuf();

// end main
return err;
}

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
int snagSave(BITMAP *bmp) {
int err = 0, i;
char filename[13];

      for (i = 0; i < 1000; i++) {
          strcpy(filename, filePrefix);
          itoa(i, &filename[strlen(filename)], 10);
          strcat(filename, ".");
          strcat(filename, fileType);

          if (!__file_exists(filename))
             break;

      } // for i

      if (i >= 1000) {
         fprintf(stderr, "You have too damn many pictures\n");
         return -1;
      }

      show_mouse(NULL);
      textout(screen, font, "Wrote: ", 0, SCREEN_H-text_height(font), 255);
      textout(screen, font, filename, text_length(font, "Wrote: "), SCREEN_H-text_height(font), 255);
      textout(screen, font, "       ", text_length(font, "Wrote: ")+text_length(font, filename), SCREEN_H-text_height(font), 255);
      err = save_bitmap(filename, bmp, pal);
//      delay(250);
//      rectfill(screen, 0, SCREEN_H-text_height(font), SCREEN_W, SCREEN_H, 0);
      blit(bitmap, screen, 0, 0, 0, 0, bitmap->w, bitmap->h);
      show_mouse(screen);

// end snagSave
return err;
}

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
int bSnag(int lx, int ty, int rx, int by) {
int err = 0, x, y;
BITMAP *bmp;

   minx = rx;
   miny = by;
   maxx = lx;
   maxy = ty;

   for (y = ty; y <= by; y++) {
//       printf("Y!\n");

//      x = lx;
//      while ((x<=lx) && !getpixel(bitmap, x, y))
//            x++;
      for (x = lx; (x <= rx) && !_getpixel(bitmap, x, y); x++)
         ;
          // _getpixel is faster, pretty sure it can't fail in this circumstance
      if (_getpixel(bitmap, x, y) && x < minx)
         minx = x;

      for (x = rx; (x >= lx) && !_getpixel(bitmap, x, y); x--)
          ;
      if (_getpixel(bitmap, x, y) && x > maxx)
         maxx = x;
   } // for y (row count)

   for (x = lx; x <= rx; x++) {
//       printf("X!\n");
       for (y = ty; (y <= by) && !_getpixel(bitmap, x, y); y++)
           ;
       if (_getpixel(bitmap, x, y) && y < miny)
          miny = y;

       for (y = by; (y >= ty) && !_getpixel(bitmap, x, y); y--)
           ;
       if (_getpixel(bitmap, x, y) && y > maxy)
          maxy = y;
   } // for x (column count)

//   fprintf(stderr, "min; max -=- (%i, %i); (%i, %i); (lx %i, ty %i):(rx %i, by %i)\n", minx, miny, maxx, maxy, lx, ty, rx, by);

   bmp = create_sub_bitmap(bitmap, minx, miny, (maxx-minx)+1, (maxy-miny)+1);
            // They're width + 1 and height + 1 because the points are inclusive
   if (!bmp) {
      fprintf(stderr, "Not enough memory for clipping bitmap\n");
   }
   blit(bitmap, bmp, minx, miny, 0, 0, bmp->w, bmp->h);
   err = snagSave(bmp);
   destroy_bitmap(bmp);

// end bSnag
return err;
}

///////////////////////////////////////////////////////////////
int boxGrab(int x, int y) { // Returns non-zero on error
int err = 0, lx, rx, ty, by;

   show_mouse(NULL);
   blit(bitmap, double_buffer, 0, 0, 0, 0, bitmap->w, bitmap->h);
   ff_drawDB();

   drawing_mode(DRAW_MODE_XOR, NULL, 0, 0);

   while (mouse_b && !key[KEY_ESC]) {
         ff_eraseScreen();

      if (x > mouse_x) {
         lx = mouse_x;
         rx = x;
      } else {
        lx = x;
        rx = mouse_x;
      }

      if (y > mouse_y) {
         ty = mouse_y;
         by = y;
      } else {
        ty = y;
        by = mouse_y;
      }

         ff_pushDrawn(lx, ty, (rx-lx)+1, (by-ty)+1);
         rect(double_buffer, lx, ty, rx, by, 255);

         ff_updateScreen();
   }


   solid_mode();
   ff_eraseScreen(); // Remove that nasty rectangle
   ff_drawDB();
   show_mouse(screen);

   // The co-ords will always be within the bitmap, becasue of the set_mouse_range up top

   if (key[KEY_ESC]) // Cancel the grab
      return 0;

   err = bSnag(lx, ty, rx, by);

// end boxGrab
return err;
}

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
void cgProcess(int x, int y) {

   if (_getpixel(bitmap, x, y) && !processed[pIndex(x, y)]) {
      if (x < minx)
         minx = x;
      else if (x > maxx)   // In original specs, I specifically noted not to put an else
                           // here... why?
           maxx = x;

      if (y < miny)
         miny = y;
      else if (y > maxy)
           maxy = y;

      processed[pIndex(x, y)] = 1;

      cgProcess(x-1, y);  // Note you could do diagonals here, but I didn't because I didn't
      cgProcess(x+1, y);  // think it was worth the time. It would only matter if you had
      cgProcess(x, y-1);  // diagonals of one pixel extending the width and height... Check
      cgProcess(x, y+1);  // that if some objects don't grab correctly...

      // Actually,I changed my mind... speed wasn't a concern at all, so I put in the
      // diagonal checks...
      cgProcess(x-1, y-1); // Remove these if speed is a concern
      cgProcess(x+1, y-1);
      cgProcess(x-1, y+1);
      cgProcess(x+1, y+1);

   }

// end cgProcess
}

///////////////////////////////////////////////////////////////
int clickGrab(int x, int y) {
int err = 0;
BITMAP *bmp;

   memset(processed, 0, sizeof(char) * (bitmap->w * bitmap->h));
                // Clear the array of processed pixels

   minx = maxx = x;
   miny = maxy = y;

   cgProcess(x, y);

   bmp = create_sub_bitmap(bitmap, minx, miny, (maxx-minx)+1, (maxy-miny)+1);
            // They're width + 1 and height + 1 because the points are inclusive
   if (!bmp) {
      fprintf(stderr, "Not enough memory for clipping bitmap\n");
   }
   blit(bitmap, bmp, minx, miny, 0, 0, bmp->w, bmp->h);
   err = snagSave(bmp);
   destroy_bitmap(bmp);

// end clickGrab
return err;
}

